@@ -1,5 +1,27 @@
 # -*- coding: utf-8 -*-
 
+"""
+Copyright (c) 2018 Keijack Wu
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in all
+copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+SOFTWARE.
+"""
+
 import atexit
 import json
 import os
@@ -30,7 +52,7 @@ except NameError:
     # python 3 does no longer support long method, use int instead
     long = int
 
-_logger = get_logger("EurekaClient")
+_logger = get_logger("eureka_client")
 
 """
 Status of instances
@@ -131,19 +153,21 @@ class Application:
                  name="",
                  instances=None):
         self.name = name
-        self.__instances = instances if instances is not None else []
+        if isinstance(instances, list):
+            for ins in instances:
+                self.add_instance(ins)
         self.__instances_dict = {}
         self.__inst_lock = RLock()
 
     @property
     def instances(self):
         with self.__inst_lock:
-            return self.__instances
+            return list(self.__instances_dict.values())
 
     @property
     def up_instances(self):
         with self.__inst_lock:
-            return [item for item in self.__instances if item.status == INSTANCE_STATUS_UP]
+            return [item for item in self.__instances_dict.values() if item.status == INSTANCE_STATUS_UP]
 
     def get_instance(self, instance_id):
         with self.__inst_lock:
@@ -154,34 +178,28 @@ class Application:
 
     def add_instance(self, instance):
         with self.__inst_lock:
-            self.__instances.append(instance)
             self.__instances_dict[instance.instanceId] = instance
 
     def update_instance(self, instance):
         with self.__inst_lock:
             _logger.debug("update instance %s" % instance.instanceId)
-            updated = False
-            for idx in range(len(self.__instances)):
-                ele = self.__instances[idx]
-                if ele.instanceId == instance.instanceId:
-                    _logger.debug("updating index %d" % idx)
-                    self.__instances[idx] = instance
-                    updated = True
-                    break
-
-            if not updated:
-                self.add_instance(instance)
+            self.__instances_dict[instance.instanceId] = instance
 
     def remove_instance(self, instance):
         with self.__inst_lock:
-            for idx in range(len(self.__instances)):
-                ele = self.__instances[idx]
-                if ele.instanceId == instance.instanceId:
-                    del self.__instances[idx]
-                    break
             if instance.instanceId in self.__instances_dict:
                 del self.__instances_dict[instance.instanceId]
 
+    def up_instances_in_zone(self, zone):
+        with self.__inst_lock:
+            _zone = zone if zone else "defaultZone"
+            return [item for item in self.__instances_dict.values() if item.status == INSTANCE_STATUS_UP and item.zone == _zone]
+
+    def up_instances_not_in_zone(self, zone):
+        with self.__inst_lock:
+            _zone = zone if zone else "defaultZone"
+            return [item for item in self.__instances_dict.values() if item.status == INSTANCE_STATUS_UP and item.zone != _zone]
+
 
 class LeaseInfo:
 
@@ -217,7 +235,7 @@ class PortWrapper:
         self.enabled = enabled
 
 
-class Instance:
+class Instance(object):
 
     def __init__(self,
                  instanceId="",
@@ -271,6 +289,13 @@ class Instance:
         self.actionType = actionType
         self.asgName = asgName
 
+    @property
+    def zone(self):
+        if self.metadata and "zone" in self.metadata and self.metadata["zone"]:
+            return self.metadata["zone"]
+        else:
+            return "defaultZone"
+
 
 ########################## Basic functions #################################
 ####### Registry functions #########
@@ -529,13 +554,13 @@ def _current_time_millis():
     return int(time.time() * 1000)
 
 
-"""====================== Registry Client ======================================="""
+"""====================== Client ======================================="""
 
 
 class EurekaServerConf(object):
 
     def __init__(self,
-                 eureka_servers=_DEFAULT_EUREKA_SERVER_URL,
+                 eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                  eureka_domain="",
                  eureka_protocol="http",
                  eureka_basic_auth_user="",
@@ -546,7 +571,7 @@ class EurekaServerConf(object):
                  zone=""):
         self.__servers = {}
         self.region = region
-        self.zone = zone
+        self.zone = _zone = zone if zone else "defaultZone"
         if eureka_domain:
             zone_urls = get_txt_dns_record("txt.%s.%s" % (region, eureka_domain))
             for zone_url in zone_urls:
@@ -563,11 +588,11 @@ class EurekaServerConf(object):
                 self.__servers[zone_name] = [self._format_url(eureka_url.strip(), eureka_protocol, eureka_basic_auth_user,
                                                               eureka_basic_auth_password, eureka_context) for eureka_url in eureka_urls]
         else:
-            self.__servers[zone if zone else "defaultZone"] = [self._format_url(eureka_url.strip(), eureka_protocol, eureka_basic_auth_user,
-                                                                                eureka_basic_auth_password, eureka_context) for eureka_url in eureka_servers.split(",")]
+            self.__servers[_zone] = [self._format_url(eureka_url.strip(), eureka_protocol, eureka_basic_auth_user,
+                                                      eureka_basic_auth_password, eureka_context) for eureka_url in eureka_server.split(",")]
         self.__servers_not_in_zone = copy(self.__servers)
-        if zone in self.__servers_not_in_zone:
-            del self.__servers_not_in_zone[zone]
+        if _zone in self.__servers_not_in_zone:
+            del self.__servers_not_in_zone[_zone]
 
     def _format_url(self, server_url="", eureka_protocol="http",
                     eureka_basic_auth_user="",
@@ -622,19 +647,187 @@ class EurekaServerConf(object):
     def servers_not_in_zone(self):
         return self.__servers_not_in_zone
 
-    def servers_ont_in_zones(self, zones=[]):
-        svs = copy(self.servers)
-        for zone in zones:
-            if zone in svs:
-                del svs[zone]
-        return svs
 
+class EurekaClient(object):
+    """
+    Example:
+    
+    >>> client = EurekaClient(
+            eureka_server="http://my_eureka_server_peer_1/eureka/v2,http://my_eureka_server_peer_2/eureka/v2", 
+            app_name="python_module_1", 
+            instance_port=9090)
+    >>> client.start()
+    >>> result = client.do_service("APP_NAME", "/context/path", return_type="json")
 
-class EurekaClient:
-    """Eureka client for spring cloud"""
+    EIPs support: 
+
+    You can configure EIP using `eureka_availability_zones` and specify the `zone` of your instance. But please aware, that the client won't fill up the metadata atomatically, 
+    You should put it to the `metadata` when creating the object.
+
+    >>> client = EurekaClient(eureka_availability_zones={
+                "us-east-1c": "http://ec2-552-627-568-165.compute-1.amazonaws.com:7001/eureka/v2/,http://ec2-368-101-182-134.compute-1.amazonaws.com:7001/eureka/v2/",
+                "us-east-1d": "http://ec2-552-627-568-170.compute-1.amazonaws.com:7001/eureka/v2/",
+                "us-east-1e": "http://ec2-500-179-285-592.compute-1.amazonaws.com:7001/eureka/v2/"}, 
+                zone="us-east-1c",
+                app_name="python_module_1", 
+                instance_port=9090,
+                data_center_name="Amazon")
+
+    EurekaClient supports DNS discovery feature.
+
+    For instance, following is a DNS TXT record created in the DNS server that lists the set of available DNS names for a zone.
+
+    >>> txt.us-east-1.mydomaintest.netflix.net="us-east-1c.mydomaintest.netflix.net" "us-east-1d.mydomaintest.netflix.net" "us-east-1e.mydomaintest.netflix.net"
+
+    Then, you can define TXT records recursively for each zone similar to the following (if more than one hostname per zone, space delimit)
+
+    >>> txt.us-east-1c.mydomaintest.netflix.net="ec2-552-627-568-165.compute-1.amazonaws.com" "ec2-368-101-182-134.compute-1.amazonaws.com"
+    >>> txt.us-east-1d.mydomaintest.netflix.net="ec2-552-627-568-170.compute-1.amazonaws.com"
+    >>> txt.us-east-1e.mydomaintest.netflix.net="ec2-500-179-285-592.compute-1.amazonaws.com"
+
+    And then you can create the client like: 
+
+    >>> client = EurekaClient(eureka_domain="mydomaintest.netflix.net",
+                region="us-east-1",
+                zone="us-east-1c",
+                app_name="python_module_1", 
+                instance_port=9090,
+                data_center_name="Amazon")
+
+    Eureka client also supports setting up the protocol, basic authentication and context path of your eureka server. 
+
+    >>> client = EurekaClient(eureka_domain="mydomaintest.netflix.net",
+                region="us-east-1",
+                zone="us-east-1c",
+                eureka_protocol="https",
+                eureka_basic_auth_user="keijack",
+                eureka_basic_auth_password="kjauthpass",
+                eureka_context="/eureka/v2",
+                app_name="python_module_1", 
+                instance_port=9090,
+                data_center_name="Amazon")
+
+    or
+
+    >>> client = EurekaClient(eureka_server="my_eureka_server_peer_1,my_eureka_server_peer_2",
+                eureka_protocol="https",
+                eureka_basic_auth_user="keijack",
+                eureka_basic_auth_password="kjauthpass",
+                eureka_context="/eureka/v2",
+                app_name="python_module_1", 
+                instance_port=9090)
+
+    You can use `do_service`, `do_service_async`, `wall_nodes`, `wall_nodes_async` to call the remote services.
+
+    >>> res = eureka_client.do_service("OTHER-SERVICE-NAME", "/service/context/path")
+
+    >>> def success_callabck(data):
+            ...
+
+        def error_callback(error):
+            ...
+        
+        client.do_service_async("OTHER-SERVICE-NAME", "/service/context/path", on_success=success_callabck, on_error=error_callback)
+
+    >>> def walk_using_your_own_urllib(url):
+            ...
+        
+        res = client.walk_nodes("OTHER-SERVICE-NAME", "/service/context/path", walker=walk_using_your_own_urllib)
+
+    >>> client.walk_nodes("OTHER-SERVICE-NAME", "/service/context/path",
+                          walker=walk_using_your_own_urllib,
+                          on_success=success_callabck,
+                          on_error=error_callback)
+        
+
+    Attributes:
+
+    * eureka_server: The eureka server url, if you want have deploy a cluster to do the failover, use `,` to separate the urls.
+
+    * eureka_domain: The domain name when using the DNS discovery.
+
+    * region: The region when using DNS discovery.
+
+    * zone: Which zone your instances belong to, default is `defaultZone`.
+
+    * eureka_availability_zones: The zones' url configurations.
+
+    * eureka_protocol: The protocol of the eureka server, if the url include this part, this protocol will not add to the url.
+
+    * eureka_basic_auth_user: User name of the basic authentication of the eureka server, if the url include this part, this protocol will not add to the url.
+
+    * eureka_basic_auth_password: Password of the basic authentication of the eureka server, if the url include this part, this protocol will not add to the url.
+
+    * eureka_context: The context path of the eureka server, if the url include this part, this protocol will not add to the url, default is `/eureka`
+        which meets the spring-boot eureka context but not the Netflix eureka server url.
+
+    * prefer_same_zone: When set to True, will first find the eureka server in the same zone to register, and find the instances in the same zone to do
+        the service. Or it will randomly choose the eureka server  to register and instances to do the services, default is `True`.
+
+    * should_register: When set to False, will not register this instance to the eureka server, default is `True`.
+
+    * should_discover: When set to False, will not pull registry from the eureka server, default is `True`.
+
+    The following parameters all the properties of this instances, all this fields will be sent to the eureka server.
+
+    * app_name: The application name of this instance.
+
+    * instance_id: The id of this instance, if not specified, will generate one by app_name and instance_host/instance_ip and instance_port.
+
+    * instance_host： The host of this instance. 
+
+    * instance_ip: The ip of this instance. If instatnce_host and instance_ip are not specified, will try to find the ip via connection to the eureka server.
+
+    * instance_port: The port of this instance. 
+
+    * instance_unsecure_port_enabled: Set whether enable the instance's unsecure port, default is `True`.
+
+    * instance_secure_port: The secure port of this instance. 
+
+    * instance_secure_port_enabled: Set whether enable the instance's secure port, default is `False`.
+
+    * data_center_name: Accept `Netflix`, `Amazon`, `MyOwn`, default is `MyOwn`
+
+    * renewal_interval_in_secs: Will send heartbeat and pull registry in this time interval, defalut is 30 seconds
+
+    * duration_in_secs: Sets the client specified setting for eviction (e.g. how long to wait without renewal event).
+
+    * home_page_url: The home page url of this instance.
+
+    * status_page_url: The status page url of this instance.
+
+    * health_check_url: The health check url of this instance.
+
+    * secure_health_check_url: The secure health check url of this instance.
+
+    * vip_adr: The virtual ip address of this instance.
+
+    * secure_vip_addr: The secure virtual ip address of this instance.
+
+    * is_coordinating_discovery_server: Sets a flag if this instance is the same as the discovery server that is
+        return the instances. This flag is used by the discovery clients to
+        identity the discovery server which is coordinating/returning the
+        information.
+
+    * metadata: The metadata map of this instances.
+
+    * remote_regions: Will also find the services that belongs to these regions.
+
+    * ha_strategy: Specify the strategy how to choose a instance when there are more than one instanse of an App. 
+
+    """
 
     def __init__(self,
                  eureka_server=_DEFAULT_EUREKA_SERVER_URL,
+                 eureka_domain="",
+                 region="",
+                 zone="",
+                 eureka_availability_zones={},
+                 eureka_protocol="http",
+                 eureka_basic_auth_user="",
+                 eureka_basic_auth_password="",
+                 eureka_context="/eureka",
+                 prefer_same_zone=True,
                  should_register=True,
                  should_discover=True,
                  app_name="",
@@ -645,7 +838,6 @@ class EurekaClient:
                  instance_unsecure_port_enabled=True,
                  instance_secure_port=_DEFAULT_INSTNACE_SECURE_PORT,
                  instance_secure_port_enabled=False,
-                 countryId=1,  # @deprecaded
                  data_center_name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
                  renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS,
                  duration_in_secs=_DURATION_IN_SECS,
@@ -659,7 +851,6 @@ class EurekaClient:
                  metadata={},
                  remote_regions=[],
                  ha_strategy=HA_STRATEGY_RANDOM):
-        assert eureka_server is not None and eureka_server != "", "eureka server must be specified."
         assert app_name is not None and app_name != "" if should_register else True, "application name must be specified."
         assert instance_port > 0 if should_register else True, "port is unvalid"
         assert isinstance(metadata, dict), "metadata must be dict"
@@ -667,9 +858,21 @@ class EurekaClient:
                                HA_STRATEGY_OTHER] if should_discover else True, "do not support strategy %d " % ha_strategy
 
         self.__net_lock = RLock()
-        self.__eureka_servers = eureka_server.split(",")
+        self.__eureka_server_conf = EurekaServerConf(
+            eureka_server=eureka_server,
+            eureka_domain=eureka_domain,
+            eureka_protocol=eureka_protocol,
+            eureka_basic_auth_user=eureka_basic_auth_user,
+            eureka_basic_auth_password=eureka_basic_auth_password,
+            eureka_context=eureka_context,
+            eureka_availability_zones=eureka_availability_zones,
+            region=region,
+            zone=zone
+        )
+        self.__cache_eureka_url = {}
         self.__should_register = should_register
         self.__should_discover = should_discover
+        self.__prefer_same_zone = prefer_same_zone
         self.__alive = False
         self.__heartbeat_timer = Timer(renewal_interval_in_secs, self.__heartbeat)
         self.__heartbeat_timer.daemon = True
@@ -689,11 +892,13 @@ class EurekaClient:
 
         # For Registery
         if should_register:
-            self.__try_all_eureka_server(try_to_get_client_ip)
+            self.__connect_to_eureka_server(try_to_get_client_ip)
 
         mdata = {
             'management.port': str(instance_port)
         }
+        if zone and zone != "defaultZone":
+            mdata["zone"] = zone
         mdata.update(metadata)
         self.__instance = {
             'instanceId': instance_id if instance_id != "" else "%s:%s:%d" % (self.__instance_host, app_name.lower(), instance_port),
@@ -708,7 +913,7 @@ class EurekaClient:
                 '$': instance_secure_port,
                 '@enabled': str(instance_secure_port_enabled).lower()
             },
-            'countryId': countryId,
+            'countryId': 1,
             'dataCenterInfo': {
                 '@class': _DEFAULT_DATA_CENTER_INFO_CLASS,
                 'name': data_center_name
@@ -732,7 +937,6 @@ class EurekaClient:
         }
 
         # For discovery
-        self.__eureka_servers = eureka_server.split(",")
         self.__remote_regions = remote_regions if remote_regions is not None else []
         self.__applications = None
         self.__delta = None
@@ -749,6 +953,10 @@ class EurekaClient:
     def should_discover(self):
         return self.__should_discover
 
+    @property
+    def zone(self):
+        return self.__eureka_server_conf.zone
+
     @property
     def applications(self):
         with self.__application_mth_lock:
@@ -756,27 +964,72 @@ class EurekaClient:
                 self.__pull_full_registry()
             return self.__applications
 
-    def __try_all_eureka_server(self, fun):
+    def __try_eureka_server_in_cache(self, fun):
+        for z, url in self.__cache_eureka_url.items():
+            _logger.debug("try to do %s in zone[%s] using cached url %s. " % (fun.__name__, z, url))
+            fun(url)
+
+    def __try_eureka_server_in_zone(self, fun):
+        self.__try_eureka_servers_in_list(fun, self.__eureka_server_conf.servers_in_zone, self.zone)
+
+    def __try_eureka_server_not_in_zone(self, fun):
+        for zone, urls in self.__eureka_server_conf.servers_not_in_zone.items():
+            try:
+                self.__try_eureka_servers_in_list(fun, urls, zone)
+            except (http_client.HTTPError, http_client.URLError):
+                _logger.exception("error!")
+            else:
+                return
+        raise http_client.URLError("All eureka servers in all zone are down!")
+
+    def __try_eureka_server_regardless_zones(self, fun):
+        for zone, urls in self.__eureka_server_conf.servers.items():
+            try:
+                self.__try_eureka_servers_in_list(fun, urls, zone)
+            except (http_client.HTTPError, http_client.URLError):
+                _logger.exception("error!")
+            else:
+                return
+        raise http_client.URLError("All eureka servers in all zone are down!")
+
+    def __try_all_eureka_servers(self, fun):
+        if self.__prefer_same_zone:
+            try:
+                self.__try_eureka_server_in_zone(fun)
+            except (http_client.HTTPError, http_client.URLError):
+                self.__try_eureka_server_not_in_zone(fun)
+        else:
+            self.__try_eureka_server_regardless_zones(fun)
+
+    def __try_eureka_servers_in_list(self, fun, eureka_servers=[], zone="defaultZone"):
         with self.__net_lock:
-            untry_servers = self.__eureka_servers
-            tried_servers = []
             ok = False
-            while len(untry_servers) > 0:
-                url = untry_servers[0].strip()
+            _zone = zone if zone else "defaultZone"
+            for url in eureka_servers:
+                url = url.strip()
                 try:
+                    _logger.debug("try to do %s in zone[%s] using url %s. " % (fun.__name__, _zone, url))
                     fun(url)
                 except (http_client.HTTPError, http_client.URLError):
                     _logger.warn("Eureka server [%s] is down, use next url to try." % url)
-                    tried_servers.append(url)
-                    untry_servers = untry_servers[1:]
                 else:
                     ok = True
+                    self.__cache_eureka_url[_zone] = url
                     break
-            if len(tried_servers) > 0:
-                untry_servers.extend(tried_servers)
-                self.__eureka_servers = untry_servers
+
             if not ok:
-                raise http_client.URLError("All eureka servers are down!")
+                if _zone in self.__cache_eureka_url:
+                    del self.__cache_eureka_url[_zone]
+                raise http_client.URLError("All eureka servers in zone[%s] are down!" % _zone)
+
+    def __connect_to_eureka_server(self, fun):
+        if self.__cache_eureka_url:
+            try:
+                self.__try_eureka_server_in_cache(fun)
+            except (http_client.HTTPError, http_client.URLError):
+                self.__try_all_eureka_servers(fun)
+        else:
+            self.__try_all_eureka_servers(fun)
 
     @staticmethod
     def __format_url(url, host, port, defalut_ctx=""):
@@ -824,7 +1077,7 @@ class EurekaClient:
         self.__instance["lastUpdatedTimestamp"] = str(_current_time_millis())
         self.__instance["lastDirtyTimestamp"] = str(_current_time_millis())
         try:
-            self.__try_all_eureka_server(lambda url: _register(url, self.__instance))
+            self.__connect_to_eureka_server(lambda url: _register(url, self.__instance))
         except:
             self.__alive = False
             _logger.exception("register error!")
@@ -834,7 +1087,7 @@ class EurekaClient:
 
     def cancel(self):
         try:
-            self.__try_all_eureka_server(lambda url: cancel(url, self.__instance["app"], self.__instance["instanceId"]))
+            self.__connect_to_eureka_server(lambda url: cancel(url, self.__instance["app"], self.__instance["instanceId"]))
         except:
             _logger.exception("error!")
         else:
@@ -845,9 +1098,9 @@ class EurekaClient:
             self.register()
             return
         try:
-            self.__try_all_eureka_server(lambda url: status_update(url, self.__instance["app"],
-                                                                   self.__instance["instanceId"], self.__instance["lastDirtyTimestamp"],
-                                                                   status=self.__instance["status"], overriddenstatus=overridden_status))
+            self.__connect_to_eureka_server(lambda url: status_update(url, self.__instance["app"],
+                                                                      self.__instance["instanceId"], self.__instance["lastDirtyTimestamp"],
+                                                                      status=self.__instance["status"], overriddenstatus=overridden_status))
         except:
             _logger.exception("Error!")
             _logger.info("Cannot send heartbeat to server, try to register")
@@ -856,13 +1109,13 @@ class EurekaClient:
     def status_update(self, new_status):
         self.__instance["status"] = new_status
         try:
-            self.__try_all_eureka_server(lambda url: status_update(url, self.__instance["app"], self.__instance["instanceId"],
-                                                                   self.__instance["lastDirtyTimestamp"], new_status))
+            self.__connect_to_eureka_server(lambda url: status_update(url, self.__instance["app"], self.__instance["instanceId"],
+                                                                      self.__instance["lastDirtyTimestamp"], new_status))
         except:
             _logger.exception("error!")
 
     def delete_status_override(self):
-        self.__try_all_eureka_server(lambda url: delete_status_override(
+        self.__connect_to_eureka_server(lambda url: delete_status_override(
             url, self.__instance["app"], self.__instance["instanceId"], self.__instance["lastDirtyTimestamp"]))
 
     def __start_registery(self):
@@ -889,7 +1142,7 @@ class EurekaClient:
             self.__applications = get_applications(url, self.__remote_regions)
             self.__delta = self.__applications
         try:
-            self.__try_all_eureka_server(do_pull)
+            self.__connect_to_eureka_server(do_pull)
         except:
             _logger.exception("pull full registry from eureka server error!")
 
@@ -909,11 +1162,10 @@ class EurekaClient:
             if not self.__is_hash_match():
                 self.__pull_full_registry()
         try:
-            self.__try_all_eureka_server(do_fetch)
+            self.__connect_to_eureka_server(do_fetch)
         except:
             _logger.exception("fetch delta from eureka server error!")
 
-
     def __is_hash_match(self):
         app_hash = self.__get_applications_hash()
         _logger.debug("check hash, local[%s], remote[%s]" % (app_hash, self.__delta.appsHashcode))
@@ -978,7 +1230,7 @@ class EurekaClient:
                     url = url + service[1:]
                 else:
                     url = url + service
-                _logger.debug("service url::" + url)
+                _logger.debug("do service with url::" + url)
                 return walker(url)
             except (http_client.HTTPError, http_client.URLError):
                 _logger.warn("do service %s in node [%s] error, use next node." % (service, node.instanceId))
@@ -1035,11 +1287,27 @@ class EurekaClient:
         app = self.applications.get_application(application_name)
         if app is None:
             return None
+        app_ups = []
+        if self.__prefer_same_zone:
+            ups_same_zone = app.up_instances_in_zone(self.zone)
+            if ups_same_zone:
+                _logger.debug("app[%s]'s up instances in the same zone are: %s" % (application_name, str([ins.instanceId for ins in ups_same_zone])))
+                app_ups.extend(ups_same_zone)
+            ups_not_same_zones = app.up_instances_not_in_zone(self.zone)
+            if ups_not_same_zones:
+                _logger.debug("app[%s]'s up instances not in the same zone are: %s" % (application_name, str([ins.instanceId for ins in ups_not_same_zones])))
+                app_ups.extend(ups_not_same_zones)
+        else:
+            ups = app.up_instances
+            if ups:
+                _logger.debug("app[%s]'s all up instances are: %s" % (application_name, str([ins.instanceId for ins in ups])))
+                app_ups = app_ups.extend(ups)
+        
         up_instances = []
         if ignore_instance_ids is None or len(ignore_instance_ids) == 0:
-            up_instances.extend(app.up_instances)
+            up_instances.extend(app_ups)
         else:
-            for ins in app.up_instances:
+            for ins in app_ups:
                 if ins.instanceId not in ignore_instance_ids:
                     up_instances.append(ins)
 
@@ -1134,6 +1402,14 @@ __cache_clients_lock = RLock()
 
 
 def init(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
+         eureka_domain="",
+         region="",
+         zone="",
+         eureka_availability_zones={},
+         eureka_protocol="http",
+         eureka_basic_auth_user="",
+         eureka_basic_auth_password="",
+         eureka_context="eureka/v2",
          should_register=True,
          should_discover=True,
          app_name="",
@@ -1144,7 +1420,6 @@ def init(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
          instance_unsecure_port_enabled=True,
          instance_secure_port=_DEFAULT_INSTNACE_SECURE_PORT,
          instance_secure_port_enabled=False,
-         countryId=1,  # @deprecaded
          data_center_name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
          renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS,
          duration_in_secs=_DURATION_IN_SECS,
@@ -1156,14 +1431,30 @@ def init(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
          secure_vip_addr="",
          is_coordinating_discovery_server=False,
          metadata={},
-         regions=[],  # @deprecated
          remote_regions=[],
          ha_strategy=HA_STRATEGY_RANDOM):
+    """
+    Initialize an EurekaClient object and put it to cache, you can use a set of functions to do the service.
+
+    Unlike using EurekaClient class that you need to start and stop the client object by yourself, this method 
+    will start the client automatically after the object created and stop it when the programe exist.
+
+    read EurekaClient for more information for the parameters details.
+    """
     with __cache_clients_lock:
-        r = regions if isinstance(regions, list) else []
-        rr = remote_regions if isinstance(remote_regions, list) else []
-        rr = r + rr
+        if __cache_key in __cache_clients:
+            _logger.warn("A client is already running, try to stop it and start the new one!")
+            __cache_clients[__cache_key].stop()
+            del __cache_clients[__cache_key]
         client = EurekaClient(eureka_server=eureka_server,
+                              eureka_domain=eureka_domain,
+                              region=region,
+                              zone=zone,
+                              eureka_availability_zones=eureka_availability_zones,
+                              eureka_protocol=eureka_protocol,
+                              eureka_basic_auth_user=eureka_basic_auth_user,
+                              eureka_basic_auth_password=eureka_basic_auth_password,
+                              eureka_context=eureka_context,
                               should_register=should_register,
                               should_discover=should_discover,
                               app_name=app_name,
@@ -1174,7 +1465,6 @@ def init(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                               instance_unsecure_port_enabled=instance_unsecure_port_enabled,
                               instance_secure_port=instance_secure_port,
                               instance_secure_port_enabled=instance_secure_port_enabled,
-                              countryId=countryId,
                               data_center_name=data_center_name,
                               renewal_interval_in_secs=renewal_interval_in_secs,
                               duration_in_secs=duration_in_secs,
@@ -1186,7 +1476,7 @@ def init(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                               secure_vip_addr=secure_vip_addr,
                               is_coordinating_discovery_server=is_coordinating_discovery_server,
                               metadata=metadata,
-                              remote_regions=rr,
+                              remote_regions=remote_regions,
                               ha_strategy=ha_strategy)
         __cache_clients[__cache_key] = client
         client.start()
@@ -1202,7 +1492,6 @@ def init_registry_client(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                          instance_unsecure_port_enabled=True,
                          instance_secure_port=_DEFAULT_INSTNACE_SECURE_PORT,
                          instance_secure_port_enabled=False,
-                         countryId=1,  # @deprecaded
                          data_center_name=_DEFAULT_DATA_CENTER_INFO,  # Netflix, Amazon, MyOwn
                          renewal_interval_in_secs=_RENEWAL_INTERVAL_IN_SECS,
                          duration_in_secs=_DURATION_IN_SECS,
@@ -1227,7 +1516,6 @@ def init_registry_client(eureka_server=_DEFAULT_EUREKA_SERVER_URL,
                 instance_unsecure_port_enabled=instance_unsecure_port_enabled,
                 instance_secure_port=instance_secure_port,
                 instance_secure_port_enabled=instance_secure_port_enabled,
-                countryId=countryId,
                 data_center_name=data_center_name,
                 renewal_interval_in_secs=renewal_interval_in_secs,
                 duration_in_secs=duration_in_secs,
